@@ -2,14 +2,14 @@ use v5.40;
 use feature 'class';
 no warnings 'experimental::class';
 #
-class Net::BitTorrent::DHT::Peer v2.0.3 {
+class Net::BitTorrent::DHT::Peer v2.0.4 {
     field $ip     : param : reader;
     field $port   : param : reader;
     field $family : param : reader;
     method to_string () {"$ip:$port"}
 };
 #
-class Net::BitTorrent::DHT v2.0.3 {
+class Net::BitTorrent::DHT v2.0.4 {
     use Algorithm::Kademlia;
     use Net::BitTorrent::DHT::Security;
     use Net::BitTorrent::Protocol::BEP03::Bencode qw[bencode bdecode];
@@ -21,7 +21,7 @@ class Net::BitTorrent::DHT v2.0.3 {
     #
     field $node_id_bin : param : reader;
     field $port             : param : reader = 6881;
-    field $address          : param = undef;
+    field $address          : param //= undef;
     field $want_v4          : param : reader = 1;
     field $want_v6          : param : reader = 1;
     field $bep32            : param : reader = 1;
@@ -46,9 +46,36 @@ class Net::BitTorrent::DHT v2.0.3 {
     field $_ed25519_backend = ();
     field $running          = 0;
     field %_blacklist;
-    #
+    field %ip_votes;    # external_ip => count
+    field $external_ip : reader = undef;
+    field %on;
+
+    method on ( $event, $cb ) {
+        push $on{$event}->@*, $cb;
+    }
+
+    method _emit ( $event, @args ) {
+        for my $cb ( $on{$event}->@* ) {
+            try { $cb->(@args) } catch ($e) {
+                warn "[ERROR] DHT event $event failed: $e"
+            };
+        }
+    }
+
+    method set_node_id ($new_id) {
+        $node_id_bin = $new_id;
+        $routing_table_v4->set_local_id_bin($new_id);
+        $routing_table_v6->set_local_id_bin($new_id);
+    }
     ADJUST {
         $socket // die "Could not create UDP socket: $!";
+        $self->on(
+            external_ip_detected => sub ($ip) {
+                return unless $bep42;
+                my $new_id = $security->generate_node_id($ip);
+                $self->set_node_id($new_id);
+            }
+        );
         if ($bep44) {
             try {
                 require Crypt::PK::Ed25519;
@@ -123,7 +150,7 @@ class Net::BitTorrent::DHT v2.0.3 {
         }
     }
 
-    method _generate_token ( $ip, $secret = undef ) {
+    method _generate_token ( $ip, $secret //= undef ) {
         $secret //= $token_secret;
         return sha1( $ip . $secret );
     }
@@ -220,9 +247,9 @@ class Net::BitTorrent::DHT v2.0.3 {
         return ( [], [], undef );
     }
 
-    method handle_incoming () {
-        my $sender = $socket->recv( my $data, 4096 );
-        return ( [], [], undef ) unless defined $data && length $data;
+    method handle_incoming ( $data //= undef, $sender //= undef ) {
+        $sender = $socket->recv( $data, 4096 ) unless defined $data;
+        return ( [], [], undef )               unless defined $data && length $data;
         my $msg;
         try { $msg = bdecode($data) }
         catch ($e) { return ( [], [], undef ) }
@@ -291,8 +318,8 @@ class Net::BitTorrent::DHT v2.0.3 {
             my $info_hash = $a->{info_hash};
             $res->{r}{token} = $self->_generate_token($ip);
             my $peers = $peer_storage->get($info_hash);
-            if ( $peers && @$peers ) {
-                my @filtered = grep { ( $_->{ip} =~ /:/ ) ? $want_v6 : $want_v4 } @$peers;
+            if ( $peers && @{ $peers->value } ) {
+                my @filtered = grep { ( $_->{ip} =~ /:/ ) ? $want_v6 : $want_v4 } @{ $peers->value };
                 $res->{r}{values} = $self->_pack_peers_raw( \@filtered );
             }
             else {
@@ -307,23 +334,23 @@ class Net::BitTorrent::DHT v2.0.3 {
         elsif ( $q eq 'announce_peer' ) {
             my $info_hash = $a->{info_hash};
             if ( $self->_verify_token( $ip, $a->{token} ) ) {
-                my $peers    = $peer_storage->get($info_hash) // [];
+                my $peers    = $peer_storage->get($info_hash) // ();
                 my $new_peer = {
                     ip => $ip,
                     port => ( $a->{implied_port} ? $port : $a->{port} ),
                     ( $bep33 && defined $a->{seed} ? ( seed => $a->{seed} ) : () )
                 };
-                @$peers = grep { $_->{ip} ne $ip } @$peers;
-                push @$peers, $new_peer;
-                $peer_storage->put( $info_hash, $peers );
+                my @peers = grep { $_->{ip} ne $ip } @{ $peers->value } if defined $peers;
+                push @peers, $new_peer;
+                $peer_storage->put( $info_hash, \@peers );
             }
         }
         elsif ( $q eq 'scrape_peers' ) {
             if ($bep33) {
                 my $info_hash = $a->{info_hash};
                 my $peers     = $peer_storage->get($info_hash) // [];
-                my $seeders   = grep { $_->{seed} } @$peers;
-                my $leechers  = @$peers - $seeders;
+                my $seeders   = grep { $_->{seed} } @{ $peers->value };
+                my $leechers  = @{ $peers->value } - $seeders;
                 $res->{r}{sn} = $seeders;
                 $res->{r}{ln} = $leechers;
             }
@@ -337,7 +364,7 @@ class Net::BitTorrent::DHT v2.0.3 {
                 my $target = $a->{target};
                 my $data   = $data_storage->get($target);
                 if ($data) {
-                    $res->{r} = { %{ $res->{r} }, %$data };
+                    $res->{r} = { %{ $res->{r} }, %{ $data->value } };
                 }
                 else {
                     my @closest;
@@ -365,8 +392,8 @@ class Net::BitTorrent::DHT v2.0.3 {
                     $to_sign .= 'v' . length($v) . ':' . $v;
                     if ( defined $_ed25519_backend && $_ed25519_backend->( $self, $a->{sig}, $to_sign, $a->{k} ) ) {
                         my $existing = $data_storage->get($target);
-                        if ( !$existing || $a->{seq} > $existing->{seq} ) {
-                            if ( !defined $a->{cas} || ( $existing && $existing->{seq} == $a->{cas} ) ) {
+                        if ( !defined $existing || $a->{seq} > $existing->value->{seq} ) {
+                            if ( !defined $a->{cas} || ( $existing && $existing->value->{seq} == $a->{cas} ) ) {
                                 $data_storage->put(
                                     $target,
                                     {   v   => $v,
@@ -412,12 +439,14 @@ class Net::BitTorrent::DHT v2.0.3 {
                 $res->{r}{nodes6} = $v6 if $v6 && $want_v6 && $bep32;
             }
         }
+        $self->_check_external_ip( $msg->{ip} ) if exists $msg->{ip};
         $self->_send_raw( bencode($res), $sender );
         return { id => $id, ip => $ip, port => $port };
     }
 
     method _handle_response ( $msg, $sender, $ip, $port ) {
-        return ( [], [], undef ) if $_blacklist{$ip};
+        $self->_check_external_ip( $msg->{ip} ) if exists $msg->{ip};
+        return ( [], [], undef )                if $_blacklist{$ip};
         my $r = $msg->{r};
         return ( [], [], undef ) unless $r && $r->{id};
         if ( $bep42 && !$security->validate_node_id( $r->{id}, $ip ) ) {
@@ -550,6 +579,25 @@ class Net::BitTorrent::DHT v2.0.3 {
         ];
     }
 
+    method _check_external_ip ($ip_bin) {
+        if ( length($ip_bin) == 6 ) {
+            $ip_bin = substr( $ip_bin, 0, 4 );
+        }
+        elsif ( length($ip_bin) == 18 ) {
+            $ip_bin = substr( $ip_bin, 0, 16 );
+        }
+        my $ip = length($ip_bin) == 4 ? inet_ntoa($ip_bin) : length($ip_bin) == 16 ? inet_ntop( AF_INET6, $ip_bin ) : undef;
+        return unless $ip;
+        $ip_votes{$ip}++;
+        if ( $ip_votes{$ip} >= 5 ) {    # Threshold for consensus
+            if ( !defined $external_ip || $external_ip ne $ip ) {
+                $external_ip = $ip;
+                $self->_emit( 'external_ip_detected', $ip );
+            }
+            %ip_votes = ();             # Reset votes after consensus
+        }
+    }
+
     method run () {
         $running = 1;
         $self->bootstrap();
